/* Assembly implementation of strchr.
   Copyright (C) 2022 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library.  If not, see
   <https://www.gnu.org/licenses/>.  */

/*
 * ISA: LoongArch64
 * Data Model: lp64
 */

/* basic algorithm :

   +.   use ld.d and mask for the first 8 bytes or less;

   +.   build a1 with 8c with dins;

   +.   use xor from a1 and a0 to check if is found;

   +.   if (a0 - 0x0101010101010101) & (~(a0 | 0x7f7f7f7f7f7f7f7f)!= 0, a0 has
   one byte is \0, else has no \0
 */

#include <sysdep.h>
#include <sys/asm.h>

#define STRCHR  strchr

/* char * strchr (const char *s1, int c); */

LEAF(STRCHR)
	.align		4
	andi		t1, a0, 0x7
	lu12i.w		a2, 0x01010
	bstrins.d	a1, a1, 15, 8
	bstrins.d	a0, zero, 2, 0

	slli.d		t1, t1, 3
	ori		a2, a2, 0x101
	bstrins.d	a1, a1, 31, 16
	ld.d		t2, a0, 0

	li.w		t0, -1
	sub.d		t3, zero, t1
	bstrins.d	a1, a1, 63, 32
	bstrins.d	a2, a2, 63, 32

	srl.d		t0, t0, t3
	slli.d		t5, a2, 7

	maskeqz		t0, t0, t1
	or		t2, t2, t0

	nor		a3, zero, t5
	sll.d		t3, a1, t1
	xor		t4, t2, t3

	sub.d		a7, t2, a2
	nor		a6, t2, a3
	sub.d		a5, t4, a2
	nor		a4, t4, a3

	and		a6, a7, a6
	and		a5, a5, a4
	or		t0, a6, a5
	bnez		t0, L(_mc8_a)
	addi.d		a0, a0, 8
L(_aloop):
	ld.d		t4, a0, 0

	xor		t2, t4, a1
	sub.d		a7, t4, a2
	nor		a6, t4, a3
	sub.d		a5, t2, a2

	nor		a4, t2, a3
	and		a6, a7, a6
	and		a5, a5, a4
	or		a7, a6, a5
	bnez		a7, L(_mc8_a)

	ld.d		t4, a0, 8
	addi.d		a0, a0, 16
	xor		t2, t4, a1
	sub.d		a7, t4, a2
	nor		a6, t4, a3
	sub.d		a5, t2, a2

	nor		a4, t2, a3
	and		a6, a7, a6
	and		a5, a5, a4
	or		a7, a6, a5
	beqz		a7, L(_aloop)

	addi.d		a0, a0, -8
L(_mc8_a):

	ctz.d		t0, a5
	ctz.d		t2, a6

	srli.w		t0, t0, 3
	srli.w		t2, t2, 3
	sltu		t1, t2, t0
	add.d		a0, a0, t0
	masknez		a0, a0, t1
	jr		ra
END(STRCHR)

#ifdef _LIBC
libc_hidden_builtin_def (strchr)
weak_alias (strchr, index)
#endif
